unit uno_pascal_proxy;

interface

{$I uno_common.inc}

uses
  Variants
  ,SysUtils
  ,Types
  ,Classes
  ,Math

  ,uno_typelib_typeclass
  ,uno_any2
  ,uno_typelib_typedescription
  ,uno_dispatcher
  ,uno_shared
  ,uno_debug
  ;

type
//------------------------------------------------------------------------------
//  Pascal proxy fo the UNO-Interface
//------------------------------------------------------------------------------
  TUNO_Pascal_proxy = class
  protected
    fReferencesCounter: Integer;
    pTypeDescr: p_typelib_InterfaceTypeDescription;
    pUnoInterface: p_uno_Interface;
    function getProxyName: AnsiString;

    procedure AfterConstruction; override;
    procedure BeforeDestruction; override;
  public
    constructor CreateByUNOInterface( const AUnoInterface: Pointer;
                                      aTypeDescr: p_typelib_InterfaceTypeDescription);
    constructor Create(AInterfaceName: AnsiString);

    procedure acquireInterface; cdecl;
    procedure releaseInterface; cdecl;

    property UNOInterface: p_uno_Interface read pUnoInterface write pUnoInterface;
    property UNOInterfaceTypeDescription: p_typelib_InterfaceTypeDescription read pTypeDescr;
    property ProxyName: AnsiString read getProxyName;
    property ReferencesCounter: Integer read fReferencesCounter;
  end;

//------------------------------------------------------------------------------
//  Wriper
//------------------------------------------------------------------------------
  TUNO_Interface = class(TCustomVariantType)
  protected
    procedure Clear( var V: TVarData ); override;
    procedure Copy( var Dest: TVarData; const Source: TVarData;
                    const Indirect: Boolean ); override;

    function FixupIdent( const AText: string ): string;
    procedure DispInvoke( Dest: PVarData; const Source: TVarData;
                          CallDesc: PCallDesc; Params: Pointer); override;
  end;

var
  UNO_Interface: TUNO_Interface;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

implementation

uses
   uno_pascal_uno
  ,uno_Data
  ,uno_rtl_ustring
  ,VarUtils;

const
  // max params count for Dispatch method
  MAX_PARAM_COUNT = 100;
  sQueryInterface: string = 'queryinterface';

//==============================================================================
//                           TUNO_Pascal_proxy
//==============================================================================
constructor TUNO_Pascal_proxy.CreateByUNOInterface(const AUnoInterface: Pointer;
                               aTypeDescr: p_typelib_InterfaceTypeDescription);
begin
  Assert( Assigned(aTypeDescr),
          sError_PascalProxy_CantCreate_ErrorInInterfaceTypeDescription);
  pTypeDescr := aTypeDescr;
  typelib_typedescription_acquire( p_typelib_TypeDescription(pTypeDescr) );
  if Assigned(AUnoInterface) then begin
    pUnoInterface := AUnoInterface;
    acquireInterface;
  end
  else pUnoInterface := nil;
end;

constructor TUNO_Pascal_proxy.Create(AInterfaceName: AnsiString);
begin
  pTypeDescr := nil;
  UnoInterface := nil;
  // Getting TypeDescription for source interface
  GetTypeDescriptionByName( p_typelib_TypeDescription(pTypeDescr),
                            AInterfaceName );
  Assert( (Assigned(pTypeDescr)),
           sError_PascalProxy_CantCreate_ErrorInInterfaceTypeDescription);
  // try to get pUnoInterface
  if ( Assigned(Pascal_Bridge) and (getUnoToPascal_Mapping<>nil) ) then
    fUnoToPascal_Mapping.mapInterface( @fUnoToPascal_Mapping,
                                       Pointer(pUnoInterface),
                                       self, pTypeDescr);
end;

procedure TUNO_Pascal_proxy.AfterConstruction;
begin
  fReferencesCounter := 1;
  Assert( Assigned(Pascal_Bridge),
          sError_PascalProxy_CantCreateWithoutBridge );
  Pascal_Bridge.RegNewProxy(self);
end;

procedure TUNO_Pascal_proxy.BeforeDestruction;
begin
  typelib_typedescription_release( p_typelib_TypeDescription(pTypeDescr) );
  releaseInterface;
  if Assigned(Pascal_Bridge) then
    Pascal_Bridge.UnRegProxy(self);
end;

procedure TUNO_Pascal_proxy.acquireInterface; cdecl;
begin
  Inc(fReferencesCounter);
  pUnoInterface^.acquire(pUnoInterface);
end;

procedure TUNO_Pascal_proxy.releaseInterface; cdecl;
begin
  Dec(fReferencesCounter);
  pUnoInterface^.release(pUnoInterface);
end;

function TUNO_Pascal_proxy.getProxyName: AnsiString;
begin
  Result := '';
  if Assigned(pTypeDescr) then
    Result := rtl_uString_To_String( pTypeDescr^.aBase.pTypeName );
end;

//------------------------------------------------------------------------------
function TUNO_Interface.FixupIdent(const AText: string): string;
begin
  Result := Uppercase(AText);
end;

procedure TUNO_Interface.Clear(var V: TVarData);
begin
  with V do begin
    if (TUNO_Pascal_proxy(VPointer).ReferencesCounter = 1) then
      FreeAndNil(TUNO_Pascal_proxy(VPointer))
    else begin
      TUNO_Pascal_proxy(VPointer).releaseInterface;
      VPointer := nil;
    end;
    VType := varEmpty;
  end;
end;

procedure TUNO_Interface.Copy( var Dest: TVarData; const Source: TVarData;
                              const Indirect: Boolean );
begin
  if Indirect and VarDataIsByRef(Source) then
    VarDataCopyNoInd(Dest, Source)
  else with Dest do begin
    VType := Source.VType;
    VPointer := Source.VPointer;
    TUNO_Pascal_proxy(Source.VPointer).acquireInterface;
  end;
end;

procedure TUNO_Interface.DispInvoke( Dest: PVarData;
                                    const Source: TVarData;
                                    CallDesc: PCallDesc;
                                    Params: Pointer);
const
  CArgTypeMask    = $7F;
  CArgByRef       = $80;

var
  LArgType: Integer;
  LArgByRef: Boolean;
  LParamPtr: Pointer;

  UNOParamArray: array of Pointer;

  curMember: p_typelib_InterfaceMemberTypeDescription;
  curMethod: p_typelib_InterfaceMethodTypeDescription;
  curMethodParam: p_typelib_MethodParameter;

  curAttribute: p_typelib_InterfaceAttributeTypeDescription;

  ArrayOfReferences: array[0..MAX_PARAM_COUNT] of Pointer;
  tmpIndex: Integer;

  InPascalParamsStack: TMemoryStream;

  // reconvert Pascal params to UNO params
  //   AParam       - Pointer on Pascal param
  //   LArgType     - Type of Variant
  //   LArgByRef    - flag. True if Pascal param pass as Reference
  //   pParamTD     - UNO param type description
  //   pOutUNOParam - pointer on conversion to UNO param
  procedure ParseParam( AParam: Pointer;
                        LArgType: Integer;
                        LArgByRef: Boolean;
                        pParamTD: p_typelib_MethodParameter;
                        var pOutUNOParam: Pointer);
  var
    tmpParamPtr: Pointer;

    tmpUNOParam_ptr: Pointer;
    tmpUNOParam_size: Integer;
    tmpUNOParam_Alig_Size: Integer;
    tmpDest: Pointer;

    tmpSize: Integer;
  begin
    // get pointer on param
    if LArgByRef
      then tmpParamPtr := Pointer(AParam^)
      else tmpParamPtr := AParam;

    if LArgType=varOleStr then begin
      {WideString was passed, not a PWideChar!}
      if Length(PWideString(tmpParamPtr)^)>0
        then tmpParamPtr := @(PWideString(tmpParamPtr)^[1])
        else tmpParamPtr := nil;
    end;

    // do "complite" always (we do not known who send data)
    typelib_typedescription_complete(@pParamTD^.pTypeRef^.pType);
    tmpUNOParam_Alig_Size := AlignSize(pParamTD^.pTypeRef^.pType);

    if not NeedNoCopy(pParamTD^.pTypeRef) then begin
      // increase stack
      InPascalParamsStack.Size := InPascalParamsStack.Size+tmpUNOParam_Alig_Size;
      // get pointer on data begin
      tmpDest := Pointer( Integer(InPascalParamsStack.Memory) + InPascalParamsStack.Position );
      InPascalParamsStack.Position := InPascalParamsStack.Position+tmpUNOParam_Alig_Size;
    end else tmpDest := nil;

    //   in   out    result
    //   0     0     Error
    //   0     1     out -> get mem
    //   1     0     in  -> send (remember about NeedPassByRef)
    //   1     1     in/out -> as in
    if (pParamTD.bIn)
      then PascalParam2UNOParam( tmpParamPtr, LArgType, pParamTD^.pTypeRef, tmpDest, tmpSize);

    tmpUNOParam_ptr := tmpDest;

    if NeedPassByRef(pParamTD^.pTypeRef) then begin
      ArrayOfReferences[tmpIndex] := tmpUNOParam_ptr;
      pOutUNOParam := @ArrayOfReferences[tmpIndex];
      inc(tmpIndex);
    end else
      pOutUNOParam := tmpUNOParam_ptr;

    // next param
    if (not LArgByRef) and (LArgType = varVariant) then begin
      if ( not (PVarData(LParamPtr)^.VType = varString) )
        then Inc(Integer(LParamPtr), SizeOf(TVarData) - SizeOf(Pointer))
        else if (CVarTypeToElementInfo[LArgType].Size = 8)
          then Inc(Integer(LParamPtr), 8 - SizeOf(Pointer));
    end;
    Inc(Integer(LParamPtr), SizeOf(Pointer));
  end;

  procedure ParseOutParams( pUNOParam: Pointer;
                            pUNOParamTD: p_typelib_MethodParameter;
                            LArgType: Integer;
                            LArgByRef: Boolean;
                            var APascalParam: Pointer
                            );
  var
    tmpParamPtr: Pointer;

    tmpUNOParam_ptr: Pointer;
    tmpUNOParam_size: Integer;
    tmpUNOParam_Alig_Size: Integer;
    tmpDest: Pointer;

    tmpSize: Integer;
  begin

    if curMethodParam^.bOut then begin
      // get pointer on param
      if LArgByRef
        then tmpParamPtr := Pointer(APascalParam^)
        else tmpParamPtr := APascalParam;

      if LArgType=varOleStr then begin
        {WideString was passed, not a PWideChar!}
        if Length(PWideString(tmpParamPtr)^)>0
          then tmpParamPtr := @(PWideString(tmpParamPtr)^[1])
          else tmpParamPtr := nil;
      end;

      tmpUNOParam_Alig_Size := AlignSize(pUNOParamTD^.pTypeRef^.pType);

      if not NeedPassByRef(pUNOParamTD^.pTypeRef)
        then UnoParam2PascalParam( pUNOParam, pUNOParamTD^.pTypeRef, tmpParamPtr, PVariant(tmpParamPtr)^)
        else UnoParam2PascalParam( PPointer(pUNOParam)^, pUNOParamTD^.pTypeRef, tmpParamPtr, PVariant(tmpParamPtr)^);

    end;

    DestroyData_and_FreeMem(pUNOParam, pUNOParamTD^.pTypeRef);

    // next param
    if (not LArgByRef) and (LArgType = varVariant) then begin
      if ( not (PVarData(LParamPtr)^.VType = varString) )
        then Inc(Integer(LParamPtr), SizeOf(TVarData) - SizeOf(Pointer))
        else if (CVarTypeToElementInfo[LArgType].Size = 8)
          then Inc(Integer(LParamPtr), 8 - SizeOf(Pointer));
    end;
    Inc(Integer(LParamPtr), SizeOf(Pointer));
  end;

var
  I, j, LArgCount: Integer;
  LIdent: string;
  // needed as bibbs in ConvertUnoParamToPascalParam
  tmpPtr: Pointer;
  tmpVar: Variant;

  pMemberParams: Pointer;
  ppMemberReturn: PPointer;
  fMemberReturnSize: Integer;
  pReturnTypeDescr: p_typelib_TypeDescription;
  tmpTD : p_typelib_TypeDescriptionReference;

  DispatchException: uno_Any;
  pDispatchException: p_uno_Any;
  ExceptionWasRaised: Boolean;
  ExceptionMessage: String;
begin
  // Grab the identifier
  LArgCount := CallDesc^.ArgCount;
  LIdent := FixupIdent(String(PChar(@CallDesc^.ArgTypes[LArgCount])));

  // Parse the arguments
  LParamPtr := Params;
  with TUNO_Pascal_proxy(Source.VPointer) do begin
    // get Member type description on calling essence (method/attribute)
    curMember := GetMemberTypeDescByName(pTypeDescr, LIdent);
    // check existence essence
    Assert( Assigned(curMember),
            Format( sError_PascalProxy_DispInvoke_NonExistentMember,
                    [ProxyName, LIdent]) );

    // method or attribute ?
    case ( curMember^.aBase.eTypeClass ) of
      typelib_TypeClass_INTERFACE_METHOD: begin
        curMethod := nil;
        TYPELIB_DANGER_GET( p_typelib_TypeDescription(curMethod),
                            curMember^.aBase.pWeakRef);
        // check sending params count
        Assert( Assigned(curMethod) and ( curMethod^.nParams = LArgCount ),
                sError_PascalProxy_DispInvoke_NotEqualParamCount);
        tmpIndex := 0;
        // pars in memory and check types
        if (LArgCount > 0) then begin
          curMethodParam := curMethod^.pParams;
{$IFDEF FPC}
          Inc(curMethodParam, LArgCount-1);
{$ENDIF}
          SetLength(UNOParamArray, LArgCount);
          // Create memory manager for output params
          InPascalParamsStack := TMemoryStream.Create;
          for i := 0 to LArgCount - 1 do
            DebugDump('DispInvoke: arg '+IntToStr(i)+' type is '+IntToStr(CallDesc^.ArgTypes[i]));
          for i := 0 to LArgCount - 1 do begin
{$IFDEF FPC}
            j := LArgCount-1-i;
{$ELSE}
            j := i;
{$ENDIF}
            LArgType := CallDesc^.ArgTypes[i] and CArgTypeMask;
            LArgByRef := (CallDesc^.ArgTypes[i] and CArgByRef) <> 0;
            ParseParam( LParamPtr,
                        LArgType,
                        LArgByRef,
                        curMethodParam,
                        UNOParamArray[j]);
{$IFDEF FPC}
            Dec(curMethodParam);
{$ELSE}
            Inc(curMethodParam);
{$ENDIF}
          end;
        end else InPascalParamsStack := nil;

        DebugDump('DispInvoke: params were handled');
        // Proxy is mapping? if not -> mapping.
        if (not Assigned(pUnoInterface)) then
          getPascalToUno_Mapping^.mapInterface( getPascalToUno_Mapping,
                                                Pointer(pUnoInterface),
                                                Source.VPointer, pTypeDescr);

        // get output param type descr
        pReturnTypeDescr := nil;
        typelib_typedescriptionreference_getDescription( pReturnTypeDescr,
                                                 curMethod.pReturnTypeRef );

        //   pReturnTypeDescr    Dest    Result
        //         1              1        Ok
        //         1              0        Ok
        //         0              0        Ok
        //         0              1       Error
        Assert(Assigned(pReturnTypeDescr) or (not Assigned(Dest)), sError_MethodHasNoReturn);

        // Getting mem for result
        case pReturnTypeDescr^.eTypeClass of
          typelib_TypeClass_VOID: ppMemberReturn := nil;
          else begin
            // allingment on 4 and get memory
            fMemberReturnSize := AlignSize(pReturnTypeDescr);
            GetMem(ppMemberReturn, fMemberReturnSize);
            ppMemberReturn^ := nil;
          end;
        end;
        DebugDump('DispInvoke: return pointer is '+IntToHex(Cardinal(ppMemberReturn), 8));

        // create in params pointer
        if ( Length(UNOParamArray) > 0 )
          then pMemberParams := @UNOParamArray[0]
          else pMemberParams := nil;

      end;

      typelib_TypeClass_INTERFACE_ATTRIBUTE: begin
//        typelib_typedescription_release(p_typelib_TypeDescription(curAttribute));
      end;
    end;

    // call to UNO
    pDispatchException := @DispatchException;
    DebugDump('DispInvoke: prepare to call method '+LIdent);
    pUnoInterface^.pDispatcher( pUnoInterface,
                                p_typelib_TypeDescription(curMember),
                                ppMemberReturn,
                                pMemberParams,
                                @pDispatchException );
    DebugDump('DispInvoke: method was called');

    // analysis "flying"
    ExceptionWasRaised := Assigned(pDispatchException);
    // process Result
    if (pReturnTypeDescr^.eTypeClass<>typelib_TypeClass_VOID) then begin
      if Assigned(Dest) then begin
//        if compare_rtl_uString_AnsiString( curMember^.pMemberName, sQueryInterface ) then
        DebugDump('DispInvoke: translate result');
        DebugDump('DispInvoke: return pointer is '+IntToHex(Cardinal(ppMemberReturn), 8));
        DebugDump('DispInvoke: dest is 0x'+IntToHex(Cardinal(Dest), 8));

  //      if not NeedPassByRef(pUNOParamTD^.pTypeRef)
  //        then UnoParam2PascalParam( pUNOParam, pUNOParamTD^.pTypeRef, tmpParamPtr, PVariant(tmpParamPtr)^)
  //        else UnoParam2PascalParam( PPointer(pUNOParam)^, pUNOParamTD^.pTypeRef, tmpParamPtr, PVariant(tmpParamPtr)^);


 //       if (pReturnTypeDescr^.eTypeClass = typelib_TypeClass_ANY) then begin
        if (not NeedPassByRef(p_typelib_TypeDescriptionReference(pReturnTypeDescr))) then begin
          DebugDump('DispInvoke: result is a variant');
          UnoParam2PascalParam( ppMemberReturn,
                                p_typelib_TypeDescriptionReference(pReturnTypeDescr),
                                tmpPtr,
                                Variant(Dest^))
        end else begin
          DebugDump('DispInvoke: result is not a variant');
          UnoParam2PascalParam( ppMemberReturn^,
                                p_typelib_TypeDescriptionReference(pReturnTypeDescr),
                                tmpPtr,
                                Variant(Dest^));
        end;
        DebugDump('DispInvoke: result was translated');
      end;

      // free memory
      if not HasWrapper((EndTypeClass(pReturnTypeDescr,ppMemberReturn))) then
          uno_destructData(ppMemberReturn, pReturnTypeDescr, nil);
//        if (EndTypeClass(pReturnTypeDescr,ppMemberReturn) <> typelib_TypeClass_INTERFACE) then
//          DestroyData_and_FreeMem(Pointer(ppMemberReturn), pReturnTypeDescr.pWeakRef);
    end;

    if ( curMember^.aBase.eTypeClass = typelib_TypeClass_INTERFACE_METHOD) then begin
      if ( LArgCount > 0) then begin
        DebugDump('DispInvoke: translate returned parameters');

        curMethodParam := curMethod^.pParams;
{$IFDEF FPC}
        Inc(curMethodParam, LArgCount-1);
{$ENDIF}
        // process Out and InOut params
        LParamPtr := Params;
        for i := 0 to LArgCount - 1 do begin
{$IFDEF FPC}
          j := LArgCount-1-i;
{$ELSE}
          j := i;
{$ENDIF}
          LArgType := CallDesc^.ArgTypes[i] and CArgTypeMask;
          LArgByRef := (CallDesc^.ArgTypes[i] and CArgByRef) <> 0;
          ParseOutParams( UNOParamArray[j],
                          curMethodParam,
                          LArgType,
                          LArgByRef,
                          LParamPtr);
{$IFDEF FPC}
          Dec(curMethodParam);
{$ELSE}
          Inc(curMethodParam);
{$ENDIF}
        end;

      end;
      typelib_typedescription_release(pReturnTypeDescr);
      TYPELIB_DANGER_RELEASE( p_typelib_TypeDescription(curMethod) );

      if Assigned(InPascalParamsStack)
        then InPascalParamsStack.Free;
    end;

  end;

  if Assigned(ppMemberReturn) then
    FreeMem(ppMemberReturn);
  FreeMemberTypeDesc(curMember);

  if ExceptionWasRaised then begin
    ExceptionMessage := rtl_uString_to_String(prtl_uString((p_uno_exception(pDispatchException^.pData)^.Mess)));
    tmpTD := typelib_static_type_getByTypeClass(typelib_TypeClass_ANY)^;
    uno_destructData(Pointer(pDispatchException), tmpTD.pType, nil);
    typelib_typedescriptionreference_release(tmpTD);
    DebugDump('DispInvoke: exception was raised during method execution ('+ExceptionMessage+')');
    raise TUNO_Exception.Create(ExceptionMessage);
  end;

// mast be released
// + pReturnTypeDescr
// + curMember: p_typelib_InterfaceMemberTypeDescription;
// + curMethod: p_typelib_InterfaceMethodTypeDescription;
// + curMethodParam: p_typelib_MethodParameter;
// + curAttribute: p_typelib_InterfaceAttributeTypeDescription;
end;

initialization
  UNO_Interface := TUNO_Interface.Create;

finalization
  if Assigned(UNO_Interface) then begin
    UNO_Interface.Free;
//    FreeAndNil(UNO_Interface);
  end;

end.
